//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "PingUnit.h"
#include <winsock2.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmPing *frmPing;

typedef struct tagIPINFO
{
	unsigned char Ttl; 			// Time To Live
	unsigned char Tos; 			// Type Of Service
	unsigned char IPFlags; 		// IP flags
	unsigned char OptSize; 		// Size of options data
	unsigned char FAR *Options; // Options data buffer
}IPINFO, *PIPINFO;

typedef struct tagICMPECHO
{
	unsigned long Source; 		// Source address
	unsigned long Status; 		// IP status
	unsigned long RTTime; 		// Round trip time in milliseconds
	unsigned short DataSize; 	// Reply data size
	unsigned short Reserved; 	// Unknown
	void FAR *pData; 			// Reply data buffer
	IPINFO ipInfo; 				// Reply options
}ICMPECHO, *PICMPECHO;


// ICMP.DLL Export Function Pointers
HANDLE (WINAPI *pIcmpCreateFile)(VOID);
BOOL (WINAPI *pIcmpCloseHandle)(HANDLE);
DWORD (WINAPI *pIcmpSendEcho)(HANDLE,DWORD,LPVOID,WORD,PIPINFO,LPVOID,DWORD,DWORD);

//---------------------------------------------------------------------------
__fastcall TfrmPing::TfrmPing(TComponent* Owner)
	: TForm(Owner)
{

}
//---------------------------------------------------------------------------
void __fastcall TfrmPing::btnStartClick(TObject *Sender)
{
	WSADATA wsaData; 			// WSADATA
	ICMPECHO icmpEcho; 			// ICMP Echo reply buffer
	HANDLE hndlIcmp; 			// LoadLibrary() handle to ICMP.DLL
	HANDLE hndlFile; 			// Handle for IcmpCreateFile()
	LPHOSTENT pHost; 			// Pointer to host entry structure
	struct in_addr iaDest; 		// Internet address structure
	IPINFO ipInfo; 				// IP Options structure
	DWORD dwRet; 				// DWORD return code

	lvReply->Clear();

	hndlIcmp = LoadLibrary("ICMP.DLL");
	if (hndlIcmp == NULL)
	{
		ShowMessage("Error loading ICMP.DLL");
		return;
	}
	// Retrieve ICMP function pointers
	pIcmpCreateFile = (HANDLE (WINAPI *)(void))	GetProcAddress(hndlIcmp,"IcmpCreateFile");
	pIcmpCloseHandle = (BOOL (WINAPI *)(HANDLE)) GetProcAddress(hndlIcmp,"IcmpCloseHandle");
	pIcmpSendEcho = (DWORD (WINAPI *)(HANDLE,DWORD,LPVOID,WORD,PIPINFO,LPVOID,DWORD,DWORD))	GetProcAddress(hndlIcmp,"IcmpSendEcho");
	// Check all the function pointers
	if (pIcmpCreateFile == NULL || pIcmpCloseHandle == NULL || pIcmpSendEcho == NULL)
	{
		ShowMessage("Error getting ICMP function address");
		FreeLibrary(hndlIcmp);
		return;
	}

	// Init WinSock
	int nRet = WSAStartup(0x0101, &wsaData );
	if (nRet)
	{
		ShowMessage("WSAStartup() error");
		WSACleanup();
		FreeLibrary(hndlIcmp);
		return;
	}
	// Check WinSock version
	if (wsaData.wVersion != 0x0101)
	{
		ShowMessage("WinSock v 1.1 not supported");
		WSACleanup();
		FreeLibrary(hndlIcmp);
		return;
	}

	// Lookup destination
	// Use inet_addr() to determine if we're dealing with a name
	// or an address
	iaDest.s_addr = inet_addr(edHost->Text.c_str());
	if (iaDest.s_addr == INADDR_NONE)
	{
		// is it a hostname?
		pHost = gethostbyname(edHost->Text.c_str());
		if (!pHost) {
			ShowMessage("Host not found");
			WSACleanup();
			FreeLibrary(hndlIcmp);
			return;
		} else {
			// Hostname was resolved, save the address */
			iaDest.s_addr = *(u_long*)pHost->h_addr;
			AnsiString txt;
			TListItem *item;
			item = lvReply->Items->Add();
			txt.sprintf("Host resolved to %s", inet_ntoa(iaDest));
			item->Caption = txt;
		}
	}


	// Get an ICMP echo request handle
	hndlFile = pIcmpCreateFile();
	for (int i=0; i<4; i++)
	{
		// Set default values
		ipInfo.Ttl = 255;
		ipInfo.Tos = 0;
		ipInfo.IPFlags = 0;
		ipInfo.OptSize = 0;
		ipInfo.Options = NULL;
		// Reqest an ICMP echo
		dwRet = pIcmpSendEcho(
			hndlFile, // Handle from IcmpCreateFile()
			iaDest.s_addr, // Destination IP address
			NULL, // Pointer to buffer to send
			0, // Size of buffer in bytes
			&ipInfo, // Request options
			&icmpEcho, // Reply buffer
			sizeof(struct tagICMPECHO),
			3000); // Time to wait in milliseconds
		Application->ProcessMessages();
		Sleep(100);
		AnsiString txt;
		TListItem *item;
		if (dwRet)
		{
			// Print the results
			iaDest.s_addr = icmpEcho.Source;
			if (icmpEcho.RTTime > 0)
				txt.sprintf("Reply: Time=%ldms TTL=%d", icmpEcho.RTTime, icmpEcho.ipInfo.Ttl);
			else
				txt.sprintf("Reply: Time<1ms TTL=%d", icmpEcho.ipInfo.Ttl);
			item = lvReply->Items->Add();
			item->Caption = txt;
		} else {
			txt.sprintf("Error: icmpEcho.Status=%ld", icmpEcho.Status);
			item = lvReply->Items->Add();
			item->Caption = txt;
			break;
		}
	}
	// Close the echo request file handle
	pIcmpCloseHandle(hndlFile);
	FreeLibrary(hndlIcmp);
	WSACleanup();
}
//---------------------------------------------------------------------------


