//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <stdio.h>
#include <Menus.hpp>
#include <IniFiles.hpp>
#include <ClipBrd.hpp>

#include "mainUnit.h"
#include "ConnectToUnit.h"
#include "Settings.h"
#include "SettingsUnit.h"
#include ".\Tools\TcpScanner\TcpScannerUnit.h"
#include ".\Tools\Ping\PingUnit.h"
#include "utils.h"
#include "AboutUnit.h"
#include "TrayIcon.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
Settings appSettings;


//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner):
	TForm(Owner),
	MRUConnections(10)
{
	trIcon = new TrayIcon(this);
	trIcon->OnLeftBtnDown = OnTrayIconLeftBtnDown;
	asLastHost = "localhost";
	iLastPort = 21;
	iHistoryMaxSize = 200;
	asLastHistoryEntry = "";
	pLogFile = NULL;
	queDisplay.reserve(500);
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::miExitClick(TObject *Sender)
{
	this->Close();
	delete trIcon;
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::miConnectToClick(TObject *Sender)
{
	frmConnectTo->asHost = asLastHost;
	frmConnectTo->iPort = iLastPort;
	frmConnectTo->ShowModal();
	asLastHost = frmConnectTo->asHost;
	iLastPort = frmConnectTo->iPort;
	if (frmConnectTo->bConnect)
	{
		ConnectTo(frmConnectTo->asHost, frmConnectTo->iPort);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::OnTrayIconLeftBtnDown(TObject *Sender)
{
	Visible = !Visible;
	if (Visible)
	{
        trIcon->ShowInTray(false);
		SetActiveWindow (Handle);
		SetForegroundWindow (Handle);
		SetWindowPos (Handle, HWND_TOP, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
		Repaint();
	}
}

void TfrmMain::SetStatus(char *lpData, ...)
{
	va_list ap;
	char buf[128]; //determines max message length
	va_start(ap, lpData);
	vsnprintf(buf, sizeof(buf)-1, lpData, ap);
	buf[sizeof(buf)-1] = NULL;
	va_end(ap);
	this->StatusBar->SimpleText = buf;
}

//---------------------------------------------------------------------------

void TfrmMain::AddBufToLog(bool bUseTimestamp, TColor color, char *Buf, int iLength)
{
	AddTextToLog(false, color, Buf);
}

void TfrmMain::AddTextToLog(bool bUseTimestamp, TColor color, AnsiString asText)
{
	if (!btnRecord->Down)
    	return;
	struct LogEntry logEntry;
	logEntry.color = color;
	logEntry.asText = asText;
	queDisplay.push_back(logEntry);

	if (pLogFile)
	{
		fwrite(asText.data(), 1, asText.Length(), pLogFile);
	}
}

void __fastcall TfrmMain::ClientSocketConnect(TObject *Sender,
      TCustomWinSocket *Socket)
{
	SetStatus("Connected to %s:%d", asLastHost.c_str(), iLastPort);
	if (appSettings.bClearWindowAtReconnect)
    {
        redMain->Clear();
	}
	if (appSettings.bKeepAlive)
	{
		int iOptval = 1;
		if(setsockopt(Socket->SocketHandle, SOL_SOCKET, SO_KEEPALIVE,
			(char *) &iOptval, sizeof(int)))
		{
			AddTextToLog(false, clWindowText, "setsockopt(SO_KEEPALIVE) failed: ");
			AddTextToLog(false, clWindowText, Utils::wstrerror(WSAGetLastError()));
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ClientSocketError(TObject *Sender,
	  TCustomWinSocket *Socket, TErrorEvent ErrorEvent, int &ErrorCode)
{
	SetStatus("%s:%d, socket error %d", asLastHost.c_str(), iLastPort,
		ErrorCode);
	ErrorCode = 0;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ClientSocketRead(TObject *Sender,
	  TCustomWinSocket *Socket)
{
	static char ReadBuf[4096];
	int size = Socket->ReceiveBuf(ReadBuf, sizeof(ReadBuf)-1);
	if (size > 0)
	{
		ReadBuf[size] = 0;
		AddBufToLog(true, clWindowText, ReadBuf, size);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ClientSocketDisconnect(TObject *Sender,
	  TCustomWinSocket *Socket)
{
	/*
	Note:
	 Non-blocking sockets do not always receive an OnRead event for the last bit
	 of data passed over the connection. When using a non-blocking socket,
	 check for any unread data in the OnDisconnect event to make sure that
	 everything is handled.
	*/
//	static char ReadBuf[4096];
//	int size = Socket->ReceiveBuf(ReadBuf, sizeof(ReadBuf)-1);
//	if (size>0)
//    {
//		AddBufToLog(true, clWindowText, ReadBuf, size);
//	}
	AddTextToLog(false, clWindowText, "\r\n");
	SetStatus("%s:%d disconnected", asLastHost.c_str(), iLastPort);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miDisconnectClick(TObject *Sender)
{
	if (ClientSocket->Active)
		ClientSocket->Close();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnSendClick(TObject *Sender)
{
	Send();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::MRUClick(TObject *Sender)
{
	TMenuItem *item = dynamic_cast<TMenuItem*>(Sender);
	AnsiString asHost;
	int iPort;
	if (Utils::StringToHostPort(StripHotkey(item->Caption), asHost, iPort) == false)
		return;
	ConnectTo(asHost, iPort);
}
//---------------------------------------------------------------------------

void TfrmMain::ConnectTo(AnsiString asHost, int iPort)
{
	ClientSocket->Active = false;
	asLastHost = asHost;
	iLastPort = iPort;
	if (Utils::IsIpValid(asHost.c_str()) && appSettings.bDetectIpAtConnect)
		ClientSocket->Address = asHost;
	else
		ClientSocket->Host = asHost;
	ClientSocket->Port = iPort;
	SetStatus("Connecting to %s:%d...", asHost.c_str(), iPort);
	ClientSocket->Active = true;

	for (int i = 0; i < miLastConnections->Count; i++)
	{
		if (StripHotkey(miLastConnections->Items[i]->Caption) == asHost + ":" + IntToStr(iPort))
		{
			delete miLastConnections->Items[i];
		}
	}
	while (miLastConnections->Count >= MRUConnections)
	{
		delete miLastConnections->Items[miLastConnections->Count -1];
	}
	TMenuItem *item = new TMenuItem(MainMenu);
	item->AutoHotkeys = maManual;
	item->Caption = asHost + ":" + IntToStr(iPort);
	miLastConnections->Insert(0, item);
	item->OnClick = MRUClick;
}

void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.Read(asConfigFile);
	this->Width = appSettings.iWidth;
	this->Height = appSettings.iHeight;
	this->Top = appSettings.iPosY;
	this->Left = appSettings.iPosX;
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.bWindowMaximized)
		this->WindowState = wsMaximized;
	btnSendOnEnter->Down = appSettings.bSendOnEnter;
	btnAutoScroll->Down = appSettings.bAutoScroll;

	TIniFile *ini = new TIniFile( asConfigFile );
	if (ini)
	{
		//bool once = false;
		for (int i = 0; i < MRUConnections; i++)
		{
			AnsiString keyName = (AnsiString)"Connection" + IntToStr(i);
			AnsiString keyValue = ini->ReadString("LastConnections", keyName, "");
			if (keyValue != "")
			{
				TMenuItem *item = new TMenuItem(MainMenu);
				item->AutoHotkeys = maManual;
				item->Caption = keyValue;
				miLastConnections->Insert(0, item);
				item->OnClick = MRUClick;

				//if (!once)
                {
					AnsiString asHost;
					int iPort;
					if (Utils::StringToHostPort(StripHotkey(item->Caption), asHost, iPort) != false)
					{
						asLastHost = asHost;
						iLastPort = iPort;
                    }
				}
			}
		}
		delete ini;
	}

	

	//auto-complete feature
	slistAutoComplete = new TStringList();
	slistAutoComplete->Sorted = true;
	slistAutoComplete->Duplicates = dupIgnore;
	AnsiString asACname = ChangeFileExt( Application->ExeName, ".autocomplete");
	if (FileExists(asACname))
	{
		slistAutoComplete->LoadFromFile(asACname);
	}


	//send history feature
	slistHistory = new TStringList();
	slistHistory->Sorted = false;
	AnsiString asHISTname = ChangeFileExt( Application->ExeName, ".history");
	iCurrentHistoryPos = 0;
	if (FileExists(asHISTname))
	{
		slistHistory->LoadFromFile(asHISTname);
		iCurrentHistoryPos = slistHistory->Count;
		if (iCurrentHistoryPos)
			asLastHistoryEntry = slistHistory->Strings[iCurrentHistoryPos-1];
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.bSendOnEnter = btnSendOnEnter->Down;
	appSettings.bAutoScroll = btnAutoScroll->Down;
	appSettings.bWindowMaximized = (this->WindowState == wsMaximized);
	if (!appSettings.bWindowMaximized)
	{
		// these values are meaningless is wnd is maximized
		appSettings.iWidth = this->Width;
		appSettings.iHeight = this->Height;
		appSettings.iPosY = this->Top;
		appSettings.iPosX = this->Left;
	}
	appSettings.Write(asConfigFile);
	TIniFile *ini = new TIniFile(asConfigFile);
	if (ini)
	{
		for (int i = miLastConnections->Count - 1; i>=0; i--)
		{
			AnsiString keyName = (AnsiString)"Connection" + IntToStr(miLastConnections->Count - 1 - i);
			ini->WriteString("LastConnections", keyName, StripHotkey(miLastConnections->Items[i]->Caption));
		}
		delete ini;
	}

	//auto-complete
	AnsiString asACname = ChangeFileExt( Application->ExeName, ".autocomplete");
	slistAutoComplete->SaveToFile(asACname);
	delete slistAutoComplete;

	//send history
	AnsiString asHISTname = ChangeFileExt( Application->ExeName, ".history");
	slistHistory->SaveToFile(asHISTname);
	delete slistHistory;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miAboutClick(TObject *Sender)
{
	AboutBox->ShowModal();	
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::edSendKeyUp(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == 8 || Key == 13 || Key == 46 || (Key>=37&&Key<=40))
	{
		return;
	}
	if (!appSettings.bAutoComplete)
	{
		return;        
	}
	// Search for unselected portion at start of text
	int p = edSend->SelStart;  // cursor position
	int len = edSend->SelLength;
	if (p + len != edSend->Text.Length())
		return;
	AnsiString theText = edSend->Text.SubString(0, p);	// Excludes searched portion
	// Match entered text portion
	if (!appSettings.bAutoCompleteCaseSensitive)
	{
		for (int i = 0; i<slistAutoComplete->Count; i++)
		{
			// Keep case of listed item
			if (slistAutoComplete->Strings[i].UpperCase().Pos(theText.UpperCase()) == 1)
			{
				edSend->Text = slistAutoComplete->Strings[i];
				edSend->SelStart = p;
				edSend->SelLength = edSend->Text.Length() - edSend->SelStart;
				break;
			}
		}
	}
	else
	{
		for (int i = 0; i<slistAutoComplete->Count; i++)
		{
			// Keep case of listed item
			if (slistAutoComplete->Strings[i].Pos(theText) == 1)
			{
				edSend->Text = slistAutoComplete->Strings[i];
				edSend->SelStart = p;
				edSend->SelLength = edSend->Text.Length() - edSend->SelStart;
				break;
			}
		}
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::edSendKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	// Search for unselected portion at start of text
	int p = edSend->SelStart;  // cursor position
	AnsiString theText = edSend->Text.SubString(0, p);	// Excludes searched portion

	if (Key == VK_RETURN)
	{
		if (appSettings.bSendOnEnter)
		{
			Send();
		}
	}
	else if (Key == VK_UP)			//history -> PREV
	{
		if (!appSettings.bAutoComplete || (theText == ""))
		{
			if (iCurrentHistoryPos > 0)
			{
				iCurrentHistoryPos--;
				edSend->Text = slistHistory->Strings[iCurrentHistoryPos];
			}
		}
		else
		{
			ShowAutoCompleteList(theText);
		}
		Key = 0;
	}
	else if (Key == VK_DOWN)	//history -> NEXT
	{
		if (iCurrentHistoryPos < slistHistory->Count - 1)
		{
			if (!appSettings.bAutoComplete || (theText == ""))
			{
				iCurrentHistoryPos++;
				edSend->Text = slistHistory->Strings[iCurrentHistoryPos];
			}
			else
			{
				ShowAutoCompleteList(theText);
			}
		}
		else
			edSend->Text = "";
		Key = 0;
	}	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miAppSettingsClick(TObject *Sender)
{
	frmSettings->appSettings = &appSettings;
	frmSettings->ShowModal();
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
}
//---------------------------------------------------------------------------

void TfrmMain::Send(void)
{
	AnsiString asSuffix;
	switch (appSettings.iSendSuffix)
	{
	case SENDSUFFIX_NONE:
		asSuffix = "";
		break;
	case SENDSUFFIX_CR:
		asSuffix = "\r";
		break;
	case SENDSUFFIX_LF:
		asSuffix = "\n";
		break;
	case SENDSUFFIX_CRLF:
		asSuffix = "\r\n";
		break;
	default:
		MessageBox(this->Handle, "Unhandled send suffix.",
			this->Caption.c_str(), MB_ICONSTOP);
		return;
	}
	AnsiString asText = edSend->Text + asSuffix;
	if (ClientSocket->Active)
	{
		ClientSocket->Socket->SendText(asText);
	}

	AddTextToLog(true, clInactiveCaption, asText);

	slistAutoComplete->Add(edSend->Text);

	if (edSend->Text != asLastHistoryEntry)
	{
		slistHistory->Add(edSend->Text);
		asLastHistoryEntry = edSend->Text;
		while (slistHistory->Count > iHistoryMaxSize)
		{
			slistHistory->Delete(0);
			--iCurrentHistoryPos;
		}
	}

	switch (appSettings.iAfterSend)
	{
	case AFTERSEND_NONE:
		iCurrentHistoryPos = slistHistory->Count;
		break;
	case AFTERSEND_CLEAR:
		iCurrentHistoryPos = slistHistory->Count;
		edSend->Clear();
	case AFTERSEND_HISTORY_NEXT:
		if (iCurrentHistoryPos < slistHistory->Count - 1)
		{
			edSend->Text = slistHistory->Strings[++iCurrentHistoryPos];
		}
		else
        	edSend->Clear();

	default:
		break;
	}
}
void __fastcall TfrmMain::btnSendOnEnterClick(TObject *Sender)
{
	appSettings.bSendOnEnter = btnSendOnEnter->Down;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnAutoScrollClick(TObject *Sender)
{
	appSettings.bAutoScroll = btnAutoScroll->Down;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miLogToFileClick(TObject *Sender)
{
	if (miLogToFile->Checked)
	{
		SaveDialog->DefaultExt = "log";
		SaveDialog->InitialDir = appSettings.asLogInitialDir;
		SaveDialog->FileName = FormatDateTime("yyyymmdd_hhnnss", Now()) + ".log";
		if (SaveDialog->Execute())
		{
            appSettings.asLogInitialDir = ExtractFileDir(SaveDialog->FileName);
			if (pLogFile)
				fclose(pLogFile);
			pLogFile = fopen(SaveDialog->FileName.c_str(),"ab+");
		}
	}
	else
	{
		fclose(pLogFile);
		pLogFile = NULL;
    }
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnClearClick(TObject *Sender)
{
	redMain->Clear();	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miTcpScannerClick(TObject *Sender)
{
	if (!frmTcpScanner->Visible)
	{
		frmTcpScanner->Show();
	}
	frmTcpScanner->BringToFront();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miTrayHideClick(TObject *Sender)
{
	trIcon->ShowInTray(true);
	Visible = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnDisconnectClick(TObject *Sender)
{
	if (ClientSocket->Active)
		ClientSocket->Close();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnReconnectClick(TObject *Sender)
{
	if (ClientSocket->Active)
		ClientSocket->Close();
	ConnectTo(asLastHost, iLastPort);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrLogRefreshTimer(TObject *Sender)
{
//	static __int64 sum;
//	static int cnt = 0;
//	LARGE_INTEGER m_StartCounter, m_EndCounter, Freq;
//	 __int64  m_ElapsedTime, *F;
//	if (!QueryPerformanceFrequency(&Freq)) {
//		ShowMessage("QueryPerformanceFrequency failed");  //VCL
//		return;
//	}
//	F=(__int64*)&Freq;
//	QueryPerformanceCounter (&m_StartCounter);
	if (queDisplay.empty())
		return;

	// disable screen redraws
	redMain->Lines->BeginUpdate();	

	// first save the current caret location
	int selStart = redMain->SelStart;
	int selLength = redMain->SelLength;

	// get the line number of the first visible line
	int firstLine = ::SendMessage(redMain->Handle, EM_GETFIRSTVISIBLELINE, 0 , 0);

	std::vector<LogEntry>::iterator iter;
	for(iter = queDisplay.begin(); iter != queDisplay.end(); ++iter)
	{
		redMain->SelStart = redMain->GetTextLen();
		redMain->SelLength = 0;
		redMain->SelAttributes->Color = iter->color;
		redMain->SelAttributes->Name = "Courier New";
		redMain->SelText = iter->asText;
	}
	queDisplay.clear();

//	QueryPerformanceCounter (&m_EndCounter);
//	m_ElapsedTime = (m_EndCounter.QuadPart  - m_StartCounter.QuadPart );
//
//	sum += m_ElapsedTime;
//	cnt++;
//	if (cnt > 0 && cnt%200==0)
//		ShowMessage("Avg execution time: " +
//			FloatToStrF((double)sum/((double)*F * cnt), ffGeneral, 5, 5));  //VCL
	int linebreak_len = redMain->Lines->LineBreak.Length();
	while ((unsigned int)redMain->Lines->Count > appSettings.iMaxUiLogLines)
	{
		selStart -= (redMain->Lines->Strings[0].Length() + linebreak_len);
		redMain->Lines->Delete(0);
		if (firstLine > 0)
			firstLine--;
	}
	if (selStart < 0)
	{
		selStart = 0;
		selLength = 0;
	}


	// restore the caret location
	redMain->SelStart = selStart;
	redMain->SelLength = selLength;

	if (appSettings.bAutoScroll)
	{
		redMain->SelStart = redMain->GetTextLen();
		redMain->Perform( EM_SCROLLCARET, 0 , 0 );
	}
	else
	{
		// get the current first visible line,
		// compute the number of lines to scroll, and scroll...
		int currLine = ::SendMessage(redMain->Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
		::SendMessage(redMain->Handle, EM_LINESCROLL, 0, firstLine - currLine);
	}
	redMain->Lines->EndUpdate();
}
//---------------------------------------------------------------------------

void TfrmMain::ShowAutoCompleteList(AnsiString theText)
{
	lboxAutoComplete->Clear();

	// Match entered text portion
	if (!appSettings.bAutoCompleteCaseSensitive)
	{
		for (int i = 0; i<slistAutoComplete->Count; i++)
		{
			// Keep case of listed item
			if (slistAutoComplete->Strings[i].UpperCase().Pos(theText.UpperCase()) == 1)
			{
				lboxAutoComplete->Items->Add(slistAutoComplete->Strings[i]);
			}
		}
	}
	else
	{
		for (int i = 0; i<slistAutoComplete->Count; i++)
		{
			// Keep case of listed item
			if (slistAutoComplete->Strings[i].Pos(theText) == 1)
			{
				lboxAutoComplete->Items->Add(slistAutoComplete->Strings[i]);
			}
		}
	}

	if (lboxAutoComplete->Items->Count)
	{
		lboxAutoComplete->Width = edSend->Width;
		lboxAutoComplete->Left = pnlBottom->Left + edSend->Left;
		lboxAutoComplete->Top = pnlBottom->Top + edSend->Top - lboxAutoComplete->Height;
		lboxAutoComplete->Visible = true;
		lboxAutoComplete->BringToFront();
		lboxAutoComplete->SetFocus();
		lboxAutoComplete->ItemIndex = 0;
	}
}

void __fastcall TfrmMain::lboxAutoCompleteKeyPress(TObject *Sender, char &Key)
{
	if (Key == VK_ESCAPE)
	{
		lboxAutoComplete->Visible = false;
		edSend->SetFocus();
	}
	else if (Key == VK_RETURN)
	{
		if (lboxAutoComplete->ItemIndex == -1)
			return;
		edSend->Text = lboxAutoComplete->Items->Strings[lboxAutoComplete->ItemIndex];
		lboxAutoComplete->Visible = false;
		edSend->SetFocus();
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::lboxAutoCompleteKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_DELETE)
	{
		if (lboxAutoComplete->ItemIndex == -1)
			return;
		int idx = slistAutoComplete->IndexOf(lboxAutoComplete->Items->Strings[lboxAutoComplete->ItemIndex]);
		if (idx != -1)
		{
			slistAutoComplete->Delete(idx);
			lboxAutoComplete->DeleteSelected();
			if (!lboxAutoComplete->Count)
				lboxAutoComplete->Hide();
		}
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::miCopyClick(TObject *Sender)
{
	Clipboard()->SetTextBuf( redMain->SelText.c_str() );
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miIcmpPingClick(TObject *Sender)
{
	if (!frmPing->Visible)
	{
		frmPing->SetHost(asLastHost);
		frmPing->Show();
	}
	frmPing->BringToFront();
}
//---------------------------------------------------------------------------

