//---------------------------------------------------------------------------


#pragma hdrstop

#include <SysUtils.hpp>
#include "utils.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)


bool Utils::StringToHostPort(AnsiString asString, AnsiString &asHost,
	int &iPort)
{
	int iColonPos = asString.Pos(":");
	if (iColonPos == 0)
		return false;
	asHost = asString.SubString(1, iColonPos-1);
	AnsiString asPort = asString.SubString(iColonPos + 1, asString.Length() - (iColonPos));
	iPort = StrToIntDef(asPort, -1);
	if ((iPort < 0 && asPort.Trim() != "0") || iPort > 65535)
    {
		return false;
	}
	return true;
}

/** \brief Check if ip address looks correct
	\param adres ip dotted decimal
    \return false when address is incorrect
*/
bool Utils::IsIpValid(char *ip)
{
    int iLastDot = -1, iDotCount = 0;
    bool bNonZero = false;
    char buf[64];
	for(unsigned int i = 0; i <= strlen(ip); i++)
    {
        // let's check if all entered char in entered
        // IP address are digits
        if(ip[i] == '.' || ip[i]==0)
        {
            if((int)i == iLastDot - 1)
            	return false;
            if(i - iLastDot > 4)
            	return false;
            if(iLastDot == -1)
            	iLastDot = 0;
            memcpy(buf, &ip[iLastDot], i-iLastDot);
            buf[i-iLastDot] = 0;
            if(atoi(buf) < 0 || atoi(buf) > 255)
            	return false;
            if(atoi(buf) != 0)
            	bNonZero = true;
            iLastDot = i;
            if(iDotCount++ > 3)
            	return false;
            continue;
        }
        if(isdigit(ip[i]) == 0)
        {
			return false;
        }
    }
    if(!bNonZero)
    	return false;
    return true;
}


/* Return as a string the error description for err */
char * Utils::wstrerror(LONG err)
{
	static LPVOID lpMsgBuf;

	if (lpMsgBuf)
		LocalFree(lpMsgBuf);
	FormatMessage(
	    FORMAT_MESSAGE_ALLOCATE_BUFFER |
	    FORMAT_MESSAGE_FROM_SYSTEM |
	    FORMAT_MESSAGE_IGNORE_INSERTS,
	    NULL,
	    err,		// GetLastError() does not seem to work reliably here
	    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
	    (LPTSTR) &lpMsgBuf,
	    0,
	    NULL
	);
	return (char*)lpMsgBuf;
}
